import unittest
import io

from cbm_files.data_file import DataFile


class TestDataFile(unittest.TestCase):
    def test_read(self):
        mock_file = io.BytesIO(b'LINE 1\r'
                               b'LINE 2\r\x00\x00\x00'
                               b'LINE 3\r')
        data = DataFile(mock_file)
        self.assertEqual(data.lines, [b'LINE 1', b'LINE 2', b'LINE 3'])

    def test_from_text(self):
        data = DataFile(None)
        text_in = io.StringIO('LINE 1\nLINE 2\nLINE 3\n')
        data.from_text(text_in)
        self.assertEqual(data.lines, [b'LINE 1', b'LINE 2', b'LINE 3'])

    def test_to_binary(self):
        data = DataFile(None)
        data.lines = [b'LINE 1',
                      b'LINE 2',
                      b'LINE 3']
        bin = list(data.to_binary())
        self.assertEqual(bin, [b'LINE 1\r', b'LINE 2\r', b'LINE 3\r'])

    def test_to_text(self):
        data = DataFile(None)
        data.lines = [b'LINE 1',
                      b'LINE 2',
                      b'LINE 3']
        text = list(data.to_text())
        self.assertEqual(text, ['LINE 1', 'LINE 2', 'LINE 3'])



import unittest
import io

from cbm_files.basic_file import BASICFile, InvalidFormatError


class TestBASICFile(unittest.TestCase):
    def test_read(self):
        mock_file = io.BytesIO(b'\x01\x12'
                               b'\x0d\x12\x0a\x00C\xb240703\x00'
                               b'\x14\x12d\x00\x87A\x00'
                               b'\x21\x12n\x00\x8bA\xb30\x89200\x00'
                               b'\x2a\x12x\x00\x97C,A\x00'
                               b'\x00\x00')
        prog = BASICFile(mock_file)
        self.assertEqual(list(prog.lines.keys()), [10, 100, 110, 120])

    def test_read_addr(self):
        mock_file = io.BytesIO(b'\x0d\x12\x0a\x00C\xb240703\x00'
                               b'\x14\x12d\x00\x87A\x00'
                               b'\x21\x12n\x00\x8bA\xb30\x89200\x00'
                               b'\x2a\x12x\x00\x97C,A\x00'
                               b'\x00\x00')
        prog = BASICFile(mock_file, start_addr=0x1201)
        self.assertEqual(list(prog.lines.keys()), [10, 100, 110, 120])

    def test_from_text(self):
        mock_file = io.StringIO("10 REM LINE 1\n30STOP\n")
        prog = BASICFile(None)
        prog.from_text(mock_file)
        self.assertEqual(list(prog.lines.keys()), [10, 30])
        self.assertEqual(prog.lines[30], b'\x90')

    def test_to_binary(self):
        prog = BASICFile(None)
        prog.add_line(10, 'C=40703')
        prog.add_line(100, 'READA')
        prog.add_line(110, 'IFA<0GOTO200')
        prog.add_line(120, 'POKEC,A')
        data = list(prog.to_binary(start_addr=0x1201, prepend_addr=False))
        self.assertEqual(data, [b'\x0d\x12\x0a\x00C\xb240703\x00',
                                b'\x14\x12d\x00\x87A\x00',
                                b'\x21\x12n\x00\x8bA\xb30\x89200\x00',
                                b'\x2a\x12x\x00\x97C,A\x00',
                                b'\x00\x00'])
        data = list(prog.to_binary(start_addr=0x1201))
        self.assertEqual(data[0], b'\x01\x12')

    def test_to_text(self):
        prog = BASICFile(None)
        prog.add_encoded_line(10, b'C\xb240703')
        prog.add_encoded_line(100, b'\x87A')
        prog.add_encoded_line(110, b'\x8bA\xb30\x89200')
        prog.add_encoded_line(120, b'\x97C,A')
        text = list(prog.to_text())
        self.assertEqual(text, ['10 C=40703', '100 READA', '110 IFA<0GOTO200', '120 POKEC,A'])

    def test_merge(self):
        prog1 = BASICFile(None)
        prog1.add_encoded_line(10, b'\x01')
        prog1.add_encoded_line(20, b'\x01')
        prog1.add_encoded_line(30, b'\x01')
        prog2 = BASICFile(None)
        prog2.add_encoded_line(15, b'\x02')
        prog2.add_encoded_line(20, b'\x02')
        prog2.add_encoded_line(25, b'\x02')
        prog1.merge(prog2)
        self.assertEqual(list(prog1.lines.keys()), [10, 15, 20, 25, 30])
        self.assertEqual(prog1.lines[20], b'\x02')

    def test_add_encoded_line(self):
        prog = BASICFile(None)
        prog.add_encoded_line(10, b'\x00')
        prog.add_encoded_line(100, b'\x00')
        prog.add_encoded_line(50, b'\x00')
        self.assertEqual(list(prog.lines.keys()), [10, 50, 100])

    def test_add_line(self):
        prog = BASICFile(None)
        prog.add_line(50, 'PRINT')
        self.assertEqual(prog.lines[50], b'\x99')

    def test_delete_line(self):
        prog = BASICFile(None)
        prog.add_encoded_line(10, b'\x00')
        prog.add_encoded_line(50, b'\x00')
        prog.add_encoded_line(100, b'\x00')
        prog.delete_line(50)
        self.assertEqual(list(prog.lines.keys()), [10, 100])

    def test_escape_set_name(self):
        self.assertEqual(BASICFile.escape_set_name('petscii-vic20en-uc'), 'escape-vic20')
        self.assertEqual(BASICFile.escape_set_name('dummy'), 'escape-c64')

    def test_tokenize(self):
        prog = BASICFile(None)
        self.assertEqual(prog.tokenize('PRINT"{lgrn}"'), b'\x99"\x99"')
        self.assertEqual(prog.tokenize('PRINT"~2a~ff"'), b'\x99"*\xff"')
        self.assertEqual(prog.tokenize('~99"*"'), b'\x99"*"')

    def test_expand(self):
        prog = BASICFile(None, encoding='utf8')
        self.assertEqual(prog.expand(b'\x99"\x09"', False), 'PRINT"\t"')

    def test_expand_escape(self):
        prog = BASICFile(None)
        self.assertEqual(prog.expand(b'\x99"\x09"', True), 'PRINT"{ensh}"')

    def test_renumber(self):
        prog1 = BASICFile(None)
        prog1.add_line(10, 'GOTO20')
        prog1.add_line(20, 'REM LINE 2')
        prog1.add_line(30, 'GOTO20')

        prog2 = prog1.renumber(100, 5)
        self.assertEqual(list(prog2.lines.keys()), [100, 105, 110])
        self.assertEqual(prog2.lines[100], b'\x89105')
        self.assertEqual(prog2.lines[110], b'\x89105')

        prog2 = prog1.renumber(100, 5, range_start=20)
        self.assertEqual(list(prog2.lines.keys()), [10, 100, 105])
        self.assertEqual(prog2.lines[10], b'\x89100')
        self.assertEqual(prog2.lines[105], b'\x89100')

        with self.assertRaises(ValueError):
            _ = prog1.renumber(5, 5, range_start=20)

def test_invalid_format(self):
        mock_file = io.BytesIO(b'\x01'*8)
        with self.assertRaises(ValueError):
            _ = BASICFile(mock_file)



import sys

try:
    import cbmcodecs
except ModuleNotFoundError:
    pass

from cbm_files import BASICFile


with open(sys.argv[1], 'rb') as fileh:
    prog = BASICFile(fileh)
for l in prog.to_text():
    print(l)



import struct

from .basic_file import BASICFile, InvalidFormatError
from .binary_file import BinaryFile


def ProgramFile(fileh, token_set='basic-v2', encoding='petscii-c64en-uc'):
    # read start address from file
    try:
        start_addr, = struct.unpack('<H', fileh.read(2))
    except struct.error:
        return None

    try:
        if start_addr & 0xFF == 1:
            # BASIC starts at 0x..01
            return BASICFile(fileh, start_addr, token_set=token_set, encoding=encoding)
    except InvalidFormatError:
        # not BASIC, rewind
        fileh.seek(2)

    return BinaryFile(fileh, start_addr)



from .tokensets import TokenSet, _token_set_register


_basic_v2_tokens = (
    ("END", 0x80),
    ("FOR", 0x81),
    ("NEXT", 0x82),
    ("DATA", 0x83),
    ("INPUT#", 0x84),
    ("INPUT", 0x85),
    ("DIM", 0x86),
    ("READ", 0x87),
    ("LET", 0x88),
    ("GOTO", 0x89),
    ("RUN", 0x8A),
    ("IF", 0x8B),
    ("RESTORE", 0x8C),
    ("GOSUB", 0x8D),
    ("RETURN", 0x8E),
    ("REM", 0x8F),
    ("STOP", 0x90),
    ("ON", 0x91),
    ("WAIT", 0x92),
    ("LOAD", 0x93),
    ("SAVE", 0x94),
    ("VERIFY", 0x95),
    ("DEF", 0x96),
    ("POKE", 0x97),
    ("PRINT#", 0x98),
    ("PRINT", 0x99),
    ("CONT", 0x9A),
    ("LIST", 0x9B),
    ("CLR", 0x9C),
    ("CMD", 0x9D),
    ("SYS", 0x9E),
    ("OPEN", 0x9F),
    ("CLOSE", 0xA0),
    ("GET", 0xA1),
    ("NEW", 0xA2),
    ("TAB(", 0xA3),
    ("TO", 0xA4),
    ("FN", 0xA5),
    ("SPC(", 0xA6),
    ("THEN", 0xA7),
    ("NOT", 0xA8),
    ("STEP", 0xA9),
    ("+", 0xAA),
    ("-", 0xAB),
    ("*", 0xAC),
    ("/", 0xAD),
    ("↑", 0xAE),
    ("^", 0xAE),
    ("AND", 0xAF),
    ("OR", 0xB0),
    (">", 0xB1),
    ("=", 0xB2),
    ("<", 0xB3),
    ("SGN", 0xB4),
    ("INT", 0xB5),
    ("ABS", 0xB6),
    ("USR", 0xB7),
    ("FRE", 0xB8),
    ("POS", 0xB9),
    ("SQR", 0xBA),
    ("RND", 0xBB),
    ("LOG", 0xBC),
    ("EXP", 0xBD),
    ("COS", 0xBE),
    ("SIN", 0xBF),
    ("TAN", 0xC0),
    ("ATN", 0xC1),
    ("PEEK", 0xC2),
    ("LEN", 0xC3),
    ("STR$", 0xC4),
    ("VAL", 0xC5),
    ("ASC", 0xC6),
    ("CHR$", 0xC7),
    ("LEFT$", 0xC8),
    ("RIGHT$", 0xC9),
    ("MID$", 0xCA),
    ("GO", 0xCB),
    ("~", 0xFF)
    )


class TokenSet_BASICv2(TokenSet):
    def __init__(self):
        super().__init__()
        self.add_tokens(_basic_v2_tokens)
        self.skip_tokenize_next_statement = 0x83
        self.skip_tokenize_eol = 0x8F
        self.ren_tokens = (0x89, 0x8A, 0x8D, 0x9B, 0XA7)

    def renumber_split(self, line_encoded):
        """Yield a sequence of substrings and integer line numbers."""
        next_part = bytearray()
        in_quote = False
        last_token = None
        val = None

        for b in line_encoded:
            if val is not None:
                # processing a line number
                if b >= ord('0') and b <= ord('9'):
                    val = val*10+(b-ord('0'))
                    continue
                else:
                    # line number complete
                    yield val
                    val = None

            if not in_quote:
                if b >= ord('0') and b <= ord('9') and last_token in self.ren_tokens:
                    # new line number
                    yield next_part
                    next_part = bytearray()
                    val = b-ord('0')
                    continue

                if b >= 0x80:
                    if last_token == 0xCB and b == 0xA4:
                        # convert 'GO TO' into 'GOTO'
                        last_token = 0x89
                    # ignore '-' token to handle 'LIST#-#' correctly
                    elif last_token != 0x9B or b != 0xAB:
                        last_token = b
                elif b == ord(':'):
                    # next statement
                    last_token = None

            if b == ord('"'):
                in_quote = not in_quote

            next_part.append(b)

        if val is None:
            if next_part:
                yield next_part
        else:
            yield val


_token_set_register['basic-v2'] = TokenSet_BASICv2()



from .tokensets import _token_set_register
from .tokenset_escape_vic20 import TokenSet_EscapeVIC20


_escape_c16_tokens = (
    ("{orng}", 0x81),
    ("{flon}", 0x82),
    ("{flof}", 0x84),
    ("{help}", 0x8C),
    ("{brn}", 0x95),
    ("{ylgn}", 0x96),
    ("{pink}", 0x97),
    ("{blgr}", 0x98),
    ("{lblu}", 0x99),
    ("{dblu}", 0x9A),
    ("{lgrn}", 0x9B)
    )


class TokenSet_EscapeC16(TokenSet_EscapeVIC20):
    def __init__(self):
        super().__init__()
        self.add_tokens(_escape_c16_tokens)


_token_set_register['escape-c16'] = TokenSet_EscapeC16()



import unittest
import io

from cbm_files.user_file import UserFile


class TestUserFile(unittest.TestCase):
    def test_user_file_read(self):
        mock_file = io.BytesIO(b'\x00\x06\x09'
                               b'ABCDEFGHI'
                               b'\x6f')
        user_file = UserFile(mock_file)
        start_addr, data = user_file.read()
        self.assertEqual(start_addr, 0x600)
        self.assertEqual(data, b'ABCDEFGHI')

    def test_user_file_write(self):
        mock_file = io.BytesIO()
        user_file = UserFile(mock_file, start_addr=0x500)
        user_file.write(b'MNBVCXZ')
        data = mock_file.getbuffer()
        self.assertEqual(data[:3], b'\x00\x05\x07')
        self.assertEqual(data[3:-1], b'MNBVCXZ')
        self.assertEqual(data[-1], 0x2a)



from .tokensets import _token_set_register
from .tokenset_basic_v2 import TokenSet_BASICv2


_basic_v3_5_tokens = (
    ("RGR", 0xCC),
    ("RLCR", 0xCD),
    ("RLUM", 0xCE),
    ("JOY", 0xCF),
    ("RDOT", 0xD0),
    ("DEC", 0xD1),
    ("HEX$", 0xD2),
    ("ERR$", 0xD3),
    ("INSTR", 0xD4),
    ("ELSE", 0xD5),
    ("RESUME", 0xD6),
    ("TRAP", 0xD7),
    ("TRON", 0xD8),
    ("TROFF", 0xD9),
    ("SOUND", 0xDA),
    ("VOL", 0xDB),
    ("AUTO", 0xDC),
    ("PUDEF", 0xDD),
    ("GRAPHIC", 0xDE),
    ("PAINT", 0xDF),
    ("CHAR", 0xE0),
    ("BOX", 0xE1),
    ("CIRCLE", 0xE2),
    ("GSHAPE", 0xE3),
    ("SSHAPE", 0xE4),
    ("DRAW", 0xE5),
    ("LOCATE", 0xE6),
    ("COLOR", 0xE7),
    ("SCNCLR", 0xE8),
    ("SCALE", 0xE9),
    ("HELP", 0xEA),
    ("DO", 0xEB),
    ("LOOP", 0xEC),
    ("EXIT", 0xED),
    ("DIRECTORY", 0xEE),
    ("DSAVE", 0xEF),
    ("DLOAD", 0xF0),
    ("HEADER", 0xF1),
    ("SCRATCH", 0xF2),
    ("COLLECT", 0xF3),
    ("COPY", 0xF4),
    ("RENAME", 0xF5),
    ("BACKUP", 0xF6),
    ("DELETE", 0xF7),
    ("RENUMBER", 0xF8),
    ("KEY", 0xF9),
    ("MONITOR", 0xFA),
    ("USING", 0xFB),
    ("UNTIL", 0xFC),
    ("WHILE", 0xFD)
    )


class TokenSet_BASICv3_5(TokenSet_BASICv2):
    """Tokens used by BASIC 3.5."""
    def __init__(self):
        super().__init__()
        self.add_tokens(_basic_v3_5_tokens)


_token_set_register['basic-v3.5'] = TokenSet_BASICv3_5()



import unittest

from cbm_files.tokenset_simons_basic import TokenSet_Simons_BASIC


class TestBasicV2(unittest.TestCase):
    def test_tokenize(self):
        tokenset = TokenSet_Simons_BASIC()
        self.assertEqual(tokenset.tokenize('SYS', 'ascii'), b'\x9e')
        self.assertEqual(tokenset.tokenize('COPY', 'ascii'), b'\x64\x77')
        self.assertEqual(tokenset.tokenize('LOW COL', 'ascii'), b'\x64\x76')

    def test_expand(self):
        tokenset = TokenSet_Simons_BASIC()
        self.assertEqual(tokenset.expand(b'\x9e', 'ascii'), 'SYS')
        self.assertEqual(tokenset.expand(b'\x64\x19', 'ascii'), 'MULTI')



import unittest

from cbm_files.tokenset_escape_vic20 import TokenSet_EscapeVIC20


class TestEscapeVIC20(unittest.TestCase):
    def test_tokenize(self):
        tokenset = TokenSet_EscapeVIC20()
        self.assertEqual(tokenset.tokenize('{red}', 'ascii'), b'\x1c')
        self.assertEqual(tokenset.tokenize('{home}', 'ascii'), b'\x13')

    def test_expand(self):
        tokenset = TokenSet_EscapeVIC20()
        self.assertEqual(tokenset.expand(b'\x1c', 'ascii'), '{red}')
        self.assertEqual(tokenset.expand(b'\x13', 'ascii'), '{home}')



import unittest
from unittest.mock import patch, Mock
import io

from cbm_files.program_file import ProgramFile


class TestProgramFile(unittest.TestCase):
    def test_program_file(self):
        mock_file = io.BytesIO(b'\x01\x12\x00\x00\x00')
        with patch('cbm_files.program_file.BASICFile', new=Mock):
            f = ProgramFile(mock_file)
        self.assertTrue(isinstance(f, Mock))

    def test_binary_file(self):
        mock_file = io.BytesIO(b'\x00\x20\x00\x00\x00')
        with patch('cbm_files.program_file.BinaryFile', new=Mock):
            f = ProgramFile(mock_file)
        self.assertTrue(isinstance(f, Mock))



import struct


class BinaryFile:
    def __init__(self, fileh, start_addr=None):
        if start_addr is None and fileh is not None:
            # read start address from file
            self.start_addr, = struct.unpack('<H', fileh.read(2))
        else:
            self.start_addr = start_addr
        self.data = fileh.read()

    def to_text(self):
        """Generator to format file contents as hex dump."""
        addr = self.start_addr

        data = self.data
        while data:
            current = data[:16]
            fmt = "${:04x}: "
            fmt += " {:02x}" * len(current)
            pack = "<{}B".format(len(current))
            yield fmt.format(addr, *struct.unpack(pack, current))
            addr += len(current)
            data = data[16:]



import struct


class CRCMismatchError(Exception):
    pass


class UserFile:
    def __init__(self, fileh, start_addr=None):
        self.fileh = fileh
        self.start_addr = start_addr

    def write(self, data, start_addr=None):
        """Write a data section, continuing the previous one if `start_addr` is absent."""
        if start_addr is None:
            start_addr = self.start_addr

        if start_addr is None:
            raise ValueError("No start address")
        if len(data) == 0 or len(data) > 256:
            raise ValueError("Invalid data length")

        csum = self._calc_csum(data)

        payload = struct.pack('<HB', start_addr, len(data))
        payload += data
        payload += bytes([csum])
        self.fileh.write(payload)
        self.start_addr = start_addr+len(data)

    def read(self):
        """Read the next data section."""
        header = self.fileh.read(3)
        start_addr, data_len = struct.unpack('<HB', header)
        data = self.fileh.read(data_len)
        csum = ord(self.fileh.read(1))

        calc_csum = self._calc_csum(data)
        if calc_csum != csum:
            raise CRCMismatchError("CRC mismatch, computed {}, read {}".format(calc_csum, csum))

        return start_addr, data

    @staticmethod
    def _calc_csum(data):
        csum = 0
        for b in data:
            csum += b
            if csum > 0xFF:
                csum &= 0xFF
                csum += 1
        return csum



from .tokensets import _token_set_register
from .tokenset_basic_v2 import TokenSet_BASICv2


_vic20_super_expander_jp_tokens = (
    ("HIRES", 0xCC),
    ("SOUND(", 0xCD),
    ("TEXT", 0xCE),
    ("PLOT", 0xCF),
    ("BOX", 0xD0),
    ("CIRCLE", 0xD1),
    ("PAINT", 0xD2),
    ("SETC", 0xD3),
    ("TEMPO", 0xD4),
    ("MUSIC", 0xD5),
    ("KEY", 0xD6),
    ("PIANO", 0xD7),
    ("LOCATE", 0xD8),
    ("CHAR", 0xD9),
    ("RELEASE", 0xDA),
    ("PDL", 0xDB),
    ("JOY", 0xDC),
    ("LIGHTX", 0xDD),
    ("LIGHTY", 0xDE),
    ("POINT", 0xDF),
    ("FGC", 0xE0),
    ("BGC", 0xE1),
    ("BDC", 0xE2)
    )


class TokenSet_VIC20_SuperExpander_JP(TokenSet_BASICv2):
    """Tokens used by Super Expander (VIC-1211/1211M)."""
    def __init__(self):
        super().__init__()
        self.add_tokens(_vic20_super_expander_jp_tokens)


_token_set_register['vic20-super-expander-jp'] = TokenSet_VIC20_SuperExpander_JP()



import unittest

from cbm_files.tokenset_basic_v2 import TokenSet_BASICv2


class TestBasicV2(unittest.TestCase):
    def test_tokenize(self):
        tokenset = TokenSet_BASICv2()
        self.assertEqual(tokenset.tokenize('SYS', 'ascii'), b'\x9e')
        self.assertEqual(tokenset.tokenize('sys', 'ascii'), b'\x9e')
        self.assertEqual(tokenset.tokenize('GOTO', 'ascii'), b'\x89')
        self.assertEqual(tokenset.tokenize('GO TO', 'ascii'), b'\xcb \xa4')
        self.assertEqual(tokenset.tokenize('INPUTA', 'ascii'), b'\x85A')
        self.assertEqual(tokenset.tokenize('INPUT#2', 'ascii'), b'\x842')
        self.assertEqual(tokenset.tokenize('GOSU:', 'ascii'), b'\xcbSU:')
        self.assertEqual(tokenset.tokenize('WAND128', 'ascii'), b'W\xaf128')
        self.assertEqual(tokenset.tokenize('A=RND(1)', 'ascii'), b'A\xb2\xbb(1)')
        self.assertEqual(tokenset.tokenize('3↑2', 'ascii'), b'3\xae2')
        self.assertEqual(tokenset.tokenize('3^2', 'ascii'), b'3\xae2')
        self.assertEqual(tokenset.tokenize('REM SHORT', 'ascii'), b'\x8f SHORT')
        self.assertEqual(tokenset.tokenize('DATA STOP:STOP', 'ascii'), b'\x83 STOP:\x90')
        self.assertEqual(tokenset.tokenize('REM STOP:STOP', 'ascii'), b'\x8f STOP:STOP')

    def test_expand(self):
        tokenset = TokenSet_BASICv2()
        self.assertEqual(tokenset.expand(b'\x9e', 'ascii'), 'SYS')
        self.assertEqual(tokenset.expand(b'A\xb2\xbb(1)', 'ascii'), 'A=RND(1)')
        self.assertEqual(tokenset.expand(b'3\xae2', 'ascii'), '3^2')

    def test_renumber_split(self):
        tokenset = TokenSet_BASICv2()
        # 'GOTO10:PRINT20'
        self.assertEqual(list(tokenset.renumber_split(b'\x8910:\x9920')), [b'\x89', 10, b':\x9920'])
        # 'ONAGOTO10,20'
        self.assertEqual(list(tokenset.renumber_split(b'\x91A\x8910,20')), [b'\x91A\x89', 10, b',', 20])
        # 'PRINT"<$8D>5"'
        self.assertEqual(list(tokenset.renumber_split(b'\x99"\x8D5"')), [b'\x99"\x8D5"'])
        # 'GO TO 30'
        self.assertEqual(list(tokenset.renumber_split(b'\xCB \xA4 30')), [b'\xCB \xA4 ', 30])
        # 'LIST40-50'
        self.assertEqual(list(tokenset.renumber_split(b'\x9B40\xAB50')), [b'\x9B', 40, b'\xAB', 50])
